package de.ubt.wm.sb.model;
import java.util.Arrays;
import java.util.HashSet;
import java.util.UUID;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.ubt.wm.sb.probs.Prob_Hit;
import de.ubt.wm.sb.probs.Prob_IndPlayer;


public class Player {
	
	static final Logger log4j = LogManager.getLogger(Player.class.getName());
		
	private UUID id;
	
	private Team team;
	
	private String name;
	private String dataFile;

	private Field field = null;
	
	private PlayerAction playerAction;
	
	private Prob_IndPlayer indProbs;
		
	public Player(String name, Field field) {
		super();
		this.name = name;
		this.field = field;
		this.id = UUID.randomUUID();
	}

	public Player(String name, String dataFile) {
		super();
		this.name = name;
		this.dataFile = dataFile;
		this.field = null;
		this.id = UUID.randomUUID();
	}	
	
	/**
	 *  sucht die passende Wahrscheinlichkeit zum Schlag des Spielers aus seiner Tabelle
	 *  flag = true <=> probSUCC
	 *  else probFAULT
	 * @throws Exception 
	 */
	public double getProbForHit(Hit hit, boolean succ, BallProperty strength) throws Exception{	
		HitTechnique tech = hit.getTechnique();
		Field targetField = hit.getTargetField();	
		Prob_Hit prob = indProbs.getProbHit(tech, targetField, this.field, strength);
		if (succ){
			return prob.getSuccProb();
		}else{
			return prob.getFaultProb();
		}
	}
	
	// update 14.07.2016 -> field Matrix wird benutzt
	public Field makeStep(Move m){
		int row = field.getRow();
		int column = field.getColumn();
		
		int sigma; // falls Player auf Seite von B steht, muss delta_row und delta_column umgekehrt werden
		if (field.belongsToSideA()){
			sigma = 1;			
		}else{
			sigma = -1;
		}
		
		if (row + sigma*m.getDelta_row() > 7){
			row = 7;
		}else if(row + sigma*m.getDelta_row() < 0){
			row = 0;
		}else{
			row = row + sigma*m.getDelta_row();
		}
		
		if (column + sigma*m.getDelta_column() > 5){
			column = 5;
		}else if(column + sigma*m.getDelta_column() < 0){
			column = 0;
		}else{
			column = column + sigma*m.getDelta_column();
		}
		
		return Field.fieldMatrix[row][column];
	}
	
	public boolean isPlayerActionAllowed(PlayerAction pa){		
		// falls Hit oder Move nicht erlaubt Playeraction auch nicht erlaubt
		if(!isHitAllowed(pa.getHit()))
			return false;
		if(!isMoveAllowed(pa.getMove()))
			return false;
		
		// echter Schlag und 2 Felder bewegen nicht zulaessig
		if (pa.getHit() != Hit.nohit && Move.MOVE_TWOSTEPS.contains(pa.getMove())){
			return false;
		}
		// Sprungschlag mit Bewegung in eine andere Richtung als sueden nicht zulaessig
		if (pa.getHit().isHitWithJump() && (pa.getMove() != Move.m_f && pa.getMove() != Move.stay)){
			return false;
		}
		// falls reception mit move oder defence mit move keine Bewegung erlaubt
		if ((pa.getHit().getTechnique() == HitTechnique.d_m || pa.getHit().getTechnique() == HitTechnique.r_m)
				&& pa.getMove() != Move.stay){
			log4j.warn("Keine Bewegung erlaubt bei d_m, r_m");
			return false;
		}
		// falls emergency shot mit Ball in neighbourfield, keine Bewegung erlaubt
		if (pa.getHit().getTechnique() == HitTechnique.F_E){
			if (Game.getInstance().getCurrentState().getFieldBall() != Game.getInstance().getCurrentState().getFieldOf(this)
					&& pa.getMove() != Move.stay){
				log4j.warn("Keine Bewegung erlaubt bei emergency shot wenn Ball nicht im selben Feld");
				return false;
			}
		}		
		return true;
	}
	
	
	/**
	 * baut die moeglichen Spieleraktionen zusammen
	 * @throws Exception 
	 */	
	public HashSet<PlayerAction> getPlayerActions() throws Exception {
		HashSet<PlayerAction> actions = new HashSet<PlayerAction>();
		HashSet<Hit> hits = getAvailableHits();
		HashSet<Move> moves = getAvailableMoves();

		for (Hit hit : hits) {
			for (Move move : moves) {
				// echter Schlag und 2 Felder bewegen nicht zulaessig
				if (hit != Hit.nohit && Move.MOVE_TWOSTEPS.contains(move)){
					continue;
				}
				
				// Sprungschlag mit Bewegung in eine andere Richtung als sueden nicht zulaessig
				if (hit.isHitWithJump() && (move != Move.m_f && move != Move.stay)){
					continue;
				}
				
				// -> das wird in Systemdynamik überprüft
//				// nicht auf die andere Seite laufen
//				if ((Arrays.asList(Field.AT_THE_NET_A).contains(this.field) || Arrays.asList(Field.AT_THE_NET_B).contains(this.field)) && (move == Move.m_f || move == Move.M_f))
//					continue;
				
				// falls reception mit move oder defence mit move keine Bewegung erlaubt
				if ((hit.getTechnique() == HitTechnique.d_m || hit.getTechnique() == HitTechnique.r_m)
						&& move != Move.stay){
					continue;
				}
				// falls emergency shot mit Ball in neighbourfield, keine Bewegung erlaubt
				if (hit.getTechnique() == HitTechnique.F_E){
					if (Game.getInstance().getCurrentState().getFieldBall() != Game.getInstance().getCurrentState().getFieldOf(this)
							&& move != Move.stay){
						continue;
					}
				}	
				PlayerAction pA = new PlayerAction(hit, move);
				if (isPlayerActionAllowed(pA)){
					actions.add(pA);
				}else{
					log4j.error("PlayerAction: " +  pA.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
					throw new Exception("PlayerAction: " +  pA.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
				}
			}
		}

		return actions;
	}
	
	public boolean isMoveAllowed(Move m){
		
		// Update 30.08. block immer erlaubt, nach der Systemdynamik hat aber nur ein Block am Netz eine Wirkung
		
//		if (m == Move.b){
			// ACHTUNG jeder kann Blocken, die Variable designated Blocker ist nur eine Hilfe zur Teamabstimmung
//			if (g.getCurrentState().getBlocker() != this){
//				return false;
//			}
			// ACHTUNG das sollte in SystemDynamik kontrolliert werden...
//			if (g.getCurrentState().isServingState())
//				return false;
//			if (!Arrays.asList(Field.AT_THE_NET_A).contains(field) && !Arrays.asList(Field.AT_THE_NET_B).contains(field)){
//				return false;
//			}
//		}
		
		// ACHTUNG wird in makeStep geschaut, dass keiner aus dem Spielfeld rauslaufen kann
//		if (m == Move.m_f){
//			if (Arrays.asList(Field.AT_THE_NET_A).contains(field) || Arrays.asList(Field.AT_THE_NET_B).contains(field)){
//				return false;
//			}
//		}
//		
//		if (m == Move.M_f){
//			if (Arrays.asList(Field.MIDDLE_ROW_A).contains(field) || Arrays.asList(Field.MIDDLE_ROW_B).contains(field) ||
//					Arrays.asList(Field.AT_THE_NET_A).contains(field) || Arrays.asList(Field.AT_THE_NET_B).contains(field)){
//				return false;
//			}
//		}
		
		return true;
	}
	
	public HashSet<Move> getAvailableMoves() throws Exception {
		// Update 30.08. block überall erlaubt, hat aber nur in den Felder am Netz einen Effekt auf Spielausgang
		HashSet<Move> availableMoves = Move.ALL_MOVES;
//		if ((Arrays.asList(Field.AT_THE_NET_A).contains(this.field) || Arrays.asList(Field.AT_THE_NET_B).contains(this.field)) && !Game.getInstance().getCurrentState().isServingState()){
//			availableMoves = new HashSet<Move>(Move.ALL_MOVES);
//		}
//		else {
//			availableMoves = new HashSet<Move>(Move.ALL_MOVES_WITHOUTBLOCKS);
//		}
		
		for (Move m : availableMoves){
			if (!isMoveAllowed(m)){
				log4j.error("Move: " +  m.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
				throw new Exception("Move: " +  m.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
			}
		}
		
		return availableMoves;
	}
	
	public boolean isHitAllowed(Hit h){
		Game game = Game.getInstance();
		State s = game.getCurrentState();
		
		// Aufschlag
		if (h.getTechnique() == HitTechnique.S_F || h.getTechnique() == HitTechnique.S_J){
			if (s.getCounter() != -1 || field != s.getFieldBall() || !isServerInItsServingState() 
					|| (!Arrays.asList(Field.SERVINGFIELDS_A).contains(field) && !Arrays.asList(Field.SERVINGFIELDS_B).contains(field))){
				return false;
			}
			if (team == game.getTeamA() && !Arrays.asList(Field.LAST_ROW_B).contains(h.getTargetField()) && !Arrays.asList(Field.MIDDLE_ROW_B).contains(h.getTargetField()))
				return false;
			if (team == game.getTeamB() && !Arrays.asList(Field.LAST_ROW_A).contains(h.getTargetField()) && !Arrays.asList(Field.MIDDLE_ROW_A).contains(h.getTargetField()))
				return false;
		}
		
		// Reception
		if (h.getTechnique() == HitTechnique.r){
			if (this.getTeam() == game.getTeamA() && !Field.SIDE_A.contains(h.getTargetField()))
				return false;
			if (this.getTeam() == game.getTeamB() && !Field.SIDE_B.contains(h.getTargetField()))
				return false;
//			if (s.getCounter() != -1)
//				return false;
			if (s.getFieldBall() != field)
				return false;
		}
		if (h.getTechnique() == HitTechnique.r_m){
			if (this.getTeam() == game.getTeamA() && !Field.SIDE_A.contains(h.getTargetField()))
				return false;
			if (this.getTeam() == game.getTeamB() && !Field.SIDE_B.contains(h.getTargetField()))
				return false;
//			if (s.getCounter() != -1)
//				return false;
			if (!s.getFieldBall().isNeighbourOrOwnField(field))
				return false;
		}
		
		// Defence
		if (h.getTechnique() == HitTechnique.d){
			if (this.getTeam() == game.getTeamA() && !Field.SIDE_A.contains(h.getTargetField()))
				return false;
			if (this.getTeam() == game.getTeamB() && !Field.SIDE_B.contains(h.getTargetField()))
				return false;
			if (s.getCounter() == -1 || s.getFieldBall() != field)
				return false;
		}
		if (h.getTechnique() == HitTechnique.d_m){
			if (this.getTeam() == game.getTeamA() && !Field.SIDE_A.contains(h.getTargetField()))
				return false;
			if (this.getTeam() == game.getTeamB() && !Field.SIDE_B.contains(h.getTargetField()))
				return false;
			if (s.getCounter() == -1 || !s.getFieldBall().isNeighbourOrOwnField(field))
				return false;
		}
		
		// setting
		if (h.getTechnique() == HitTechnique.s){
			// Ball muss angenommen oder verteidigt worden sein
			if (!(s.getCounter() > 0))
				return false;
			// target Feld muss auf eigener Seite sein
			if (this.getTeam() == game.getTeamA() && !Field.SIDE_A.contains(h.getTargetField()))
				return false;
			if (this.getTeam() == game.getTeamB() && !Field.SIDE_B.contains(h.getTargetField()))
				return false;
			// target Field muss Nachbarfeld sein
			if (!field.isNeighbourOrOwnField(h.getTargetField())) 
				return false;
		}
		
		// field attack
		if (h.getTechnique() == HitTechnique.F_SM){
			if (!(s.getCounter() > 1) || !s.getFieldBall().isSameFieldOrOneFieldSouthOf(field))
				return false;
			if (team == game.getTeamA() && !Arrays.asList(Field.LAST_ROW_B).contains(h.getTargetField()) && !Arrays.asList(Field.MIDDLE_ROW_B).contains(h.getTargetField())){
				return false;
			}
			if (team == game.getTeamB() && !Arrays.asList(Field.LAST_ROW_A).contains(h.getTargetField()) && !Arrays.asList(Field.MIDDLE_ROW_A).contains(h.getTargetField())){
				return false;
			}
		}
		if (h.getTechnique() == HitTechnique.F_E){
			if (!(s.getCounter() > 1) || !s.getFieldBall().isNeighbourOrOwnField(field))
				return false;
			if (team == game.getTeamA() && !Arrays.asList(Field.LAST_ROW_B).contains(h.getTargetField()) && !Arrays.asList(Field.MIDDLE_ROW_B).contains(h.getTargetField())){
				return false;
			}
			if (team == game.getTeamB() && !Arrays.asList(Field.LAST_ROW_A).contains(h.getTargetField()) && !Arrays.asList(Field.MIDDLE_ROW_A).contains(h.getTargetField())){
				return false;
			}
		}
		if (h.getTechnique() == HitTechnique.F_P){
			if (!(s.getCounter() > 0) || !s.getFieldBall().isNeighbourOrOwnField(field))
				return false;
			if (team == game.getTeamA() && !Arrays.asList(Field.INSIDE_B).contains(h.getTargetField())){
				return false;
			}
			if (team == game.getTeamB() && !Arrays.asList(Field.INSIDE_A).contains(h.getTargetField())){
				return false;
			}
		}
		
		return true;
	}
	
	public HashSet<Hit> getAvailableHits() throws Exception {
		Game game = Game.getInstance();
		
		HashSet<Hit> availableHits = new HashSet<Hit>();
		availableHits.add(Hit.nohit);		
		
		// serve
		if (isServerInItsServingState()) {
			if (isInTeamA()) {
				for (Hit h : Hit.SERVES_TEAMA){
					if (isHitAllowed(h)){
						availableHits.add(h);
					}else{
						log4j.error("Hit: " +  h.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
						throw new Exception("Hit: " +  h.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
					}
				}
			}
			if (isInTeamB()) {
				for (Hit h : Hit.SERVES_TEAMB){
					if (isHitAllowed(h)){
						availableHits.add(h);
					}else{
						log4j.error("Hit: " +  h.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
						throw new Exception("Hit: " +  h.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
					}
				}
			}
		}
		
		// reception
		if (game.getCurrentState().getCounter() == -1 && !game.getCurrentState().isServingState()){
			if (game.getCurrentState().getFieldBall() == field){
				Hit h;
				if (this.isInTeamA()){
					for (Field f : Field.INSIDE_A){
						h = new Hit(HitTechnique.r,f);
						if (isHitAllowed(h)){
							availableHits.add(h);
						}else{
							log4j.error("Hit: " +  h.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
							throw new Exception("Hit: " +  h.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
						}
					}				
				}
				if (this.isInTeamB()){
					for (Field f : Field.INSIDE_B){
						h = new Hit(HitTechnique.r,f);
						if (isHitAllowed(h)){
							availableHits.add(h);
						}else{
							log4j.error("Hit: " +  h.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
							throw new Exception("Hit: " +  h.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
						}
					}				
				}
			}
			if (game.getCurrentState().getFieldBall().isNeighbourOrOwnField(field)){
				Hit h;
				if (this.isInTeamA()){
					for (Field f : Field.INSIDE_A){
						h = new Hit(HitTechnique.r_m,f);
						if (isHitAllowed(h)){
							availableHits.add(h);
						}else{
							log4j.error("Hit: " +  h.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
							throw new Exception("Hit: " +  h.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
						}
					}				
				}
				if (this.isInTeamB()){
					for (Field f : Field.INSIDE_B){
						h = new Hit(HitTechnique.r_m,f);
						if (isHitAllowed(h)){
							availableHits.add(h);
						}else{
							log4j.error("Hit: " +  h.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
							throw new Exception("Hit: " +  h.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
						}
					}				
				}
			}
		}

		// setting (Ball muss angenommen worden sein)
		if (ballIsInMyField() && game.getCurrentState().getCounter() > 0) {
			HashSet<Field> neighbours = new HashSet<Field>(Arrays.asList(this.field.getNeighbours()));
			neighbours.add(field);
			for(Field f : neighbours){
				if (isInTeamA() && f.belongsToSideA()){
					Hit h = new Hit(HitTechnique.s, f);
					if (isHitAllowed(h)){
						availableHits.add(h);
					}else{
						log4j.error("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
						throw new Exception("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
					}
				}
				if (isInTeamB() && f.belongsToSideB()){
					Hit h = new Hit(HitTechnique.s, f);
					if (isHitAllowed(h)){
						availableHits.add(h);
					}else{
						log4j.error("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
						throw new Exception("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
					}
				}
			}
		}

		// attack
		// fuer Smash Ball muss gestellt worden sein und muss im eigenen Feld oder im Feld südlich sein
		if ((ballIsInMyField() || ballIsSouthNeighbourOfMyField()) && game.getCurrentState().getCounter() > 1) {
			if(isInTeamA()){
				for (Hit h : Hit.SMASH_TEAMA){
					if (isHitAllowed(h)){
						availableHits.add(h);
					}else{
						log4j.error("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
						throw new Exception("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
					}
				}
			}
			if(isInTeamB()){
				for (Hit h : Hit.SMASH_TEAMB){
					if (isHitAllowed(h)){
						availableHits.add(h);
					}else{
						log4j.error("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
						throw new Exception("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
					}
				}
			}
		}
		// fuer Shot Ball muss gestellt worden sein und im Nachbarfeld sein
				if (ballIsNeighbourOrInMyField() && game.getCurrentState().getCounter() > 1) {
					if(isInTeamA()){
						for (Hit h : Hit.SHOT_TEAMA){
							if (isHitAllowed(h)){
								availableHits.add(h);
							}else{
								log4j.error("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
								throw new Exception("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
							}
						}
					}
					if(isInTeamB()){
						for (Hit h : Hit.SHOT_TEAMB){
							if (isHitAllowed(h)){
								availableHits.add(h);
							}else{
								log4j.error("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
								throw new Exception("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
							}
						}
					}
				}
		// fuer Poke Ball muss angenommen worden sein
		if (ballIsInMyField() && game.getCurrentState().getCounter() > 0) {
			if(isInTeamA()){
				for (Hit h : Hit.POKE_TEAMA){
					if (isHitAllowed(h)){
						availableHits.add(h);
					}else{
						log4j.error("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
						throw new Exception("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
					}
				}
			}
			if(isInTeamB()){
				for (Hit h : Hit.POKE_TEAMB){
					if (isHitAllowed(h)){
						availableHits.add(h);
					}else{
						log4j.error("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
						throw new Exception("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
					}
				}
			}
		}

		// defence
		// && !isAtTheNet() entfernt 14.07.2016
		if (game.getCurrentState().getCounter() != -1){
			if (game.getCurrentState().getFieldBall() == field){
				Hit h;
				if (this.isInTeamA()){
					for (Field f : Field.INSIDE_A){
						h = new Hit(HitTechnique.d,f);
						if (isHitAllowed(h)){
							availableHits.add(h);
						}else{
							log4j.error("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
							throw new Exception("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
						}
					}				
				}
				if (this.isInTeamB()){
					for (Field f : Field.INSIDE_B){
						h = new Hit(HitTechnique.d,f);
						if (isHitAllowed(h)){
							availableHits.add(h);
						}else{
							log4j.error("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
							throw new Exception("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
						}
					}				
				}
			}
			if (game.getCurrentState().getFieldBall().isNeighbourOrOwnField(field)){
				Hit h;
				if (this.isInTeamA()){
					for (Field f : Field.INSIDE_A){
						h = new Hit(HitTechnique.d_m,f);
						if (isHitAllowed(h)){
							availableHits.add(h);
						}else{
							log4j.error("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
							throw new Exception("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
						}
					}				
				}
				if (this.isInTeamB()){
					for (Field f : Field.INSIDE_B){
					h = new Hit(HitTechnique.d_m,f);
					
					if (isHitAllowed(h)){
						availableHits.add(h);
					}else{
						log4j.error("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
						throw new Exception("Hit: " +  h.toString() +" is not allowed. Das sollte nicht vorkommen");
					}
					}				
				}
			}
			
		}
		
		return availableHits;
	}
	
	
	public boolean isAtTheNet(){
		if (isInTeamA() && Arrays.asList(Field.AT_THE_NET_A).contains(field)) {
			return true;
		}
		if (isInTeamB() && Arrays.asList(Field.AT_THE_NET_B).contains(field)) {
			return true;
		}
		return false;
	}
	
	public boolean ballIsOnMySide(){
		Game game = Game.getInstance();
		if (isInTeamA() && game.getCurrentState().getFieldBall().belongsToSideA()){
				return true;
			}
		if (isInTeamB() && game.getCurrentState().getFieldBall().belongsToSideB()){
				return true;
			}
		return false;		
	}
	
	public boolean ballIsInMyField(){
		Game game = Game.getInstance();
		if (this.field == game.getCurrentState().getFieldBall())
			return true;
		else
			return false;
	}
	
	public boolean ballIsSouthNeighbourOfMyField(){
		Game game = Game.getInstance();
		if (field.belongsToSideA() && game.getCurrentState().getFieldBall().belongsToSideA()
				&& game.getCurrentState().getFieldBall().getColumn() == field.getColumn() 
				&& game.getCurrentState().getFieldBall().getRow() == field.getRow()+1){
			return true;
		}
		if (field.belongsToSideB() && game.getCurrentState().getFieldBall().belongsToSideB()
				&& game.getCurrentState().getFieldBall().getColumn() == field.getColumn() 
				&& game.getCurrentState().getFieldBall().getRow() == field.getRow()-1){
			return true;
		}
		
		return false;
	}
	
	public boolean ballIsNeighbourOrInMyField(){
		Game game = Game.getInstance();
		if (this.field == game.getCurrentState().getFieldBall() || this.field.isNeighbour(game.getCurrentState().getFieldBall()))
			return true;
		else 
			return false;
	}

	
	/**
	 *  @return true if Player is Server and State is One of its Serving States
	 */	
	public boolean isServerInItsServingState(){
		Game game = Game.getInstance();
		State s = game.getCurrentState();
		if (this == s.getLastServer()){
			if (this == game.getTeamA().getPlayer1() && Arrays.asList(State.START_STATES_A1).contains(s))
				return true;
			if (this == game.getTeamA().getPlayer2() && Arrays.asList(State.START_STATES_A2).contains(s))
				return true;
			if (this == game.getTeamB().getPlayer1() && Arrays.asList(State.START_STATES_B1).contains(s))
				return true;
			if (this == game.getTeamB().getPlayer2() && Arrays.asList(State.START_STATES_B2).contains(s))
				return true;
		}
		return false;
	}
	
	public boolean madeLastContact(){
		Game game = Game.getInstance();
		if (game.getCurrentState().getCounter() != 0 && game.getCurrentState().getCounter() != -1 && game.getCurrentState().getLastContact() == this){
			return true;
		}
		else return false;
	}
	
	public boolean isInTeamA(){
		Game game = Game.getInstance();
		if (game.getTeamA() == this.team)
			return true;
		else
			return false;
		}
	
	public boolean isInTeamB(){
		Game game = Game.getInstance();
		if (game.getTeamB() == this.team)
			return true;
		else
			return false;
		}

	public UUID getId() {
		return id;
	}

	public void setId(UUID id) {
		this.id = id;
	}

	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public String getDataFile() {
		return dataFile;
	}

	public void setDataFile(String dataFile) {
		this.dataFile = dataFile;
	}

	public Field getField() {
		return field;
	}

	public void setField(Field field) {
		this.field = field;
	}

	public PlayerAction getPlayerAction() {
		return playerAction;
	}

	public void setPlayerAction(PlayerAction playeraction) {
		this.playerAction = playeraction;
	}

	public Team getTeam() {
		return team;
	}

	public void setTeam(Team team) {
		this.team = team;
	}

	public Prob_IndPlayer getIndProbs() {
		return indProbs;
	}

	public void setIndProbs(Prob_IndPlayer indProbs) {
		// original Probs kopieren und anpassen;
		this.indProbs = indProbs;
		
		indProbs.computeAggregatedProbs(this);
//		System.out.println(indProbs.printAggProbs());
		indProbs.computeModifiedProbs();
		
	}
	
	/*
	 * fuer beliebige Strategien muss Abschätzung gefunden werden
	 */	
	public double[] getMinMaxSuccFaultDev(boolean max, HashSet<HitTechnique> techs){
		double[] d;
		if (max){
			d = new double[]{0.0, 0.0, 0.0};
		}else{
			d = new double[]{1.0, 1.0, 1.0};
		}
			
		HashSet<Field> targetFields;
		HashSet<Field> playerFields;
		if (isInTeamA()){
			playerFields = Field.SIDE_A;
			targetFields = Field.SIDE_B;			
		}else{
			playerFields = Field.SIDE_B;
			targetFields = Field.SIDE_A;
		}

		for (HitTechnique tech : techs){
			for (Field playerField : playerFields){
				for (Field targetField: targetFields){
					for (BallProperty strength : Arrays.asList(BallProperty.HARD, BallProperty.NORMAL)){
						if (this.indProbs.getProbHit(tech, targetField, playerField, strength) != null){
							if (max){
								if (this.indProbs.getProbHit(tech, targetField, playerField, strength).getSuccProb() > d[0])
									d[0] = this.indProbs.getProbHit(tech, targetField, playerField, strength).getSuccProb();
								if (this.indProbs.getProbHit(tech, targetField, playerField, strength).getFaultProb() > d[1])
									d[1] = this.indProbs.getProbHit(tech, targetField, playerField, strength).getFaultProb();
								if (this.indProbs.getProbHit(tech, targetField, playerField, strength).getNeighbourProb() > d[2])
									d[2] = this.indProbs.getProbHit(tech, targetField, playerField, strength).getNeighbourProb();
							}else{
								if (this.indProbs.getProbHit(tech, targetField, playerField, strength).getSuccProb() < d[0])
									d[0] = this.indProbs.getProbHit(tech, targetField, playerField, strength).getSuccProb();
								if (this.indProbs.getProbHit(tech, targetField, playerField, strength).getFaultProb() < d[1])
									d[1] = this.indProbs.getProbHit(tech, targetField, playerField, strength).getFaultProb();
								if (this.indProbs.getProbHit(tech, targetField, playerField, strength).getNeighbourProb() < d[2])
									d[2] = this.indProbs.getProbHit(tech, targetField, playerField, strength).getNeighbourProb();
							}
						};
					}
				}
			}
		}		
		return d;
	}
		
	public String printIndProbs(){
		return this.indProbs.printAnalogToLatexFormat(this);
	}

	@Override
	public String toString() {
		return "Player of team: " + team.name +" [ name=" + name + ", field=" + field.getName() + "]";
	}
}
