package de.ubt.wm.sb.model;

import java.util.Arrays;
import java.util.HashSet;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.ubt.wm.sb.model.strategy.RealStrat;
import de.ubt.wm.sb.probs.Prob_IndPlayer;
import de.ubt.wm.sb.sim.MainSim;

public class Team {

	static final Logger log4j = LogManager.getLogger(Team.class.getName());
		
	public String name;
	
	private Player player1;	
	private Player player2;
	
	private RealStrat strat;
	private TeamAction teamAction;
	
	public Team(String name) {
		this.name = name;
	}

	public void addPlayersToTeam(Player player1, Player player2){
		this.player1 = player1;
		this.player2 = player2;
		player1.setTeam(this);
		player2.setTeam(this);
	}
	
	public boolean makesHit(){
		if (teamAction.getActionPlayer1().getHit() != Hit.nohit ||
				teamAction.getActionPlayer2().getHit() != Hit.nohit)
			return true;
		return false;
	}
	
	public boolean makesSet(){
		if (teamAction.getActionPlayer1().getHit().getTechnique() == HitTechnique.s ||
				teamAction.getActionPlayer2().getHit().getTechnique() == HitTechnique.s)
			return true;
		return false;
	}
	
	public void chooseTeamAction(double r) throws Exception {
		
		TeamAction teamAction = strat.getTeamAction(Game.getInstance().getCurrentState(), r);
		
		if (isTeamActionAllowed(teamAction)){
			setTeamAction(teamAction);
		}else{
			log4j.error("TeamAction: " + teamAction.toString() + " of team " + this.name + " not allowed in state " + Game.getInstance().getCurrentState().toString());
			throw new Exception("TeamAction: " + teamAction.toString() + " of team " + this.name + " not allowed in state " + Game.getInstance().getCurrentState().toString());
		}
	}
	
	/**
	 * nur fuer Testzwecke um zu zeigen, dass basic strategy sinn macht
	 * @param r
	 * @throws Exception
	 */
	public void chooseTeamActionAtRandom(double r) throws Exception {
		HashSet<TeamAction> allowedTeamActions = this.getAllowedTeamActions();
		TeamAction teamAction;
		if (allowedTeamActions.size()>0){
			TeamAction[] hitsArray = allowedTeamActions.toArray(new TeamAction[allowedTeamActions.size()]);
			int l = hitsArray.length;
			int n = MainSim.r.nextInt(l);
			teamAction = hitsArray[n];
		}else{
			teamAction = TeamAction.doNothing;
		}	
		setTeamAction(teamAction);
	}
	
	/**
	 * 
	 * @param tA die überpüft werden soll, ob sie erlaubt ist
	 * @return true wenn Aktion erlaubt
	 * Achtung: isTeamActionAllowed implementiert die Regeln nochmals selbst, da es wesentlich länger dauert getAllowedActions().contains(tA) auszuwerten!
	 */
	public boolean isTeamActionAllowed(TeamAction tA){		
		if (!this.player1.isPlayerActionAllowed(tA.getActionPlayer1()))
			return false;
		if (!this.player2.isPlayerActionAllowed(tA.getActionPlayer2()))
			return false;
		
		if (tA.getActionPlayer1().getHit() != Hit.nohit && tA.getActionPlayer2().getHit() != Hit.nohit){
			log4j.warn("Nur ein Spieler darf Ball schlagen!");
			return false;
		}
		
		return true;
	}

	
	/**
	 *  Baut die Menge moeglichen Teamaktionen aus den moeglichen Spieleraktionen zusammen
	 *  Dass die Regeln mit isTeamActionAllowed übereinstimmen wird in if Abfrage gewährleistet
	 * @throws Exception wenn eine Aktion in der Menge ist, die von der Funktion isTeamActionAllowed nicht erlaubt ist
	 **/
	public HashSet<TeamAction> getAllowedTeamActions() throws Exception{
		HashSet<TeamAction> teamActions = new HashSet<TeamAction>();
	
		HashSet<PlayerAction> player1Actions = player1.getPlayerActions();
		HashSet<PlayerAction> player2Actions = player2.getPlayerActions();
		
		for (PlayerAction p1Act : player1Actions){
			for (PlayerAction p2Act : player2Actions){
				
				// Aktionen bei denen beide Spieler den Ball spielen, werden rausgenommen
				if (p1Act.getHit() != Hit.nohit && p2Act.getHit() != Hit.nohit){
					continue;
				}
				
				for (Player p : Arrays.asList(player1,player2,null)){
					TeamAction tA = new TeamAction(p1Act, p2Act, p);
					if (isTeamActionAllowed(tA)){
						teamActions.add(new TeamAction(p1Act, p2Act, p));
					}else{
						log4j.error("Team Action: " +  tA.toString() +" is not allowed in state: " + Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
						throw new Exception("Team Action: " +  tA.toString() +" is not allowed in state: "+  Game.getInstance().getCurrentState().toString() +". Das sollte nicht vorkommen");
					}
				}
			}
		}	
		
		return teamActions;	
	}

	/**
	 * function calculates average skills per team for FSM, FP, SJ, SJ
	 * @return
	 */
	public String getAggregatedSkillsForSSC(){
		Prob_IndPlayer probA1 = player1.getIndProbs();
		Prob_IndPlayer probA2 = player2.getIndProbs();
		
		StringBuffer result = new StringBuffer("Aggregated Skills of Team\n");
		
		// FSM
		{
			int totalObs = probA1.F_SM.getNumberOfObservations()+probA2.F_SM.getNumberOfObservations();
			
			double fsm_succ = (probA1.F_SM.getNumberOfObservations()*probA1.F_SM.getSuccProb()
					+ probA2.F_SM.getNumberOfObservations()*probA2.F_SM.getSuccProb())/totalObs;
			double fsm_fault = (probA1.F_SM.getNumberOfObservations()*probA1.F_SM.getFaultProb()
					+ probA2.F_SM.getNumberOfObservations()*probA2.F_SM.getFaultProb())/totalObs;
			double fsm_dev = 1 - fsm_succ - fsm_fault;
	
			result.append("FSM: [succ: " + fsm_succ + ", fault: " + fsm_fault + ", dev: " + fsm_dev + " ] \n");
		}
		{
			int totalObs_fp = probA1.F_P.getNumberOfObservations()+probA2.F_P.getNumberOfObservations();
			
			double fp_succ = (probA1.F_P.getNumberOfObservations()*probA1.F_P.getSuccProb()
					+ probA2.F_P.getNumberOfObservations()*probA2.F_P.getSuccProb())/totalObs_fp;
			double fp_fault = (probA1.F_P.getNumberOfObservations()*probA1.F_P.getFaultProb()
					+ probA2.F_P.getNumberOfObservations()*probA2.F_P.getFaultProb())/totalObs_fp;
			double fp_dev = 1 - fp_succ - fp_fault;
	
			result.append("FP: [succ: " + fp_succ + ", fault: " + fp_fault + ", dev: " + fp_dev + " ] \n");
		}
		{
			int totalObs_sj = probA1.S_J.getNumberOfObservations()+probA2.S_J.getNumberOfObservations();
			
			double sj_succ = (probA1.S_J.getNumberOfObservations()*probA1.S_J.getSuccProb()
					+ probA2.S_J.getNumberOfObservations()*probA2.S_J.getSuccProb())/totalObs_sj;
			double sj_fault = (probA1.S_J.getNumberOfObservations()*probA1.S_J.getFaultProb()
					+ probA2.S_J.getNumberOfObservations()*probA2.S_J.getFaultProb())/totalObs_sj;
			double sj_dev = 1 - sj_succ - sj_fault;
	
			result.append("S_J: [succ: " + sj_succ + ", fault: " + sj_fault + ", dev: " + sj_dev + " ] \n");
		}
		{
			int totalObs_sf = probA1.S_F.getNumberOfObservations()+probA2.S_F.getNumberOfObservations();
			
			double sf_succ = (probA1.S_F.getNumberOfObservations()*probA1.S_F.getSuccProb()
					+ probA2.S_F.getNumberOfObservations()*probA2.S_F.getSuccProb())/totalObs_sf;
			double sf_fault = (probA1.S_F.getNumberOfObservations()*probA1.S_F.getFaultProb()
					+ probA2.S_F.getNumberOfObservations()*probA2.S_F.getFaultProb())/totalObs_sf;
			double sf_dev = 1 - sf_succ - sf_fault;
	
			result.append("S_F: [succ: " + sf_succ + ", fault: " + sf_fault + ", dev: " + sf_dev + " ] \n");
		}
		return result.toString();
	}
	
	public Player getPlayer1() {
		return player1;
	}

	public void setPlayer1(Player player1) {
		this.player1 = player1;
	}

	public Player getPlayer2() {
		return player2;
	}

	public void setPlayer2(Player player2) {
		this.player2 = player2;
	}

	public void setStrat(RealStrat strat) {
		this.strat = strat;
	}

	public RealStrat getStrat() {
		return strat;
	}

	public TeamAction getTeamAction() throws Exception {
		if (teamAction.getActionPlayer1().getHit().getTechnique() != player1.getPlayerAction().getHit().getTechnique() || 
				teamAction.getActionPlayer1().getHit().getTargetField() != player1.getPlayerAction().getHit().getTargetField() ||
				teamAction.getActionPlayer1().getMove() != player1.getPlayerAction().getMove() ||
				
				teamAction.getActionPlayer2().getHit().getTechnique() != player2.getPlayerAction().getHit().getTechnique() || 
				teamAction.getActionPlayer2().getHit().getTargetField() != player2.getPlayerAction().getHit().getTargetField() ||
				teamAction.getActionPlayer2().getMove() != player2.getPlayerAction().getMove()){
			throw new Exception("Team Action passt nicht zu player Actions; TeamAction: " + teamAction.toString()
			+ " Player1 Action: " + player1.getPlayerAction()
			+ " Player2 Action: " + player2.getPlayerAction());
		}
		
		
		return this.teamAction;
	}

	public void setTeamAction(TeamAction teamAction) throws Exception {
		
		if(isTeamActionAllowed(teamAction)){
			player1.setPlayerAction(teamAction.getActionPlayer1());
			player2.setPlayerAction(teamAction.getActionPlayer2());
			this.teamAction = teamAction;
		}else{
			throw new Exception("TeamAction " + teamAction.toString() + " in state " + Game.getInstance().getCurrentState().toString() 
					+ " not allowed!");
		}
	}

	public boolean isDefendingTeam() {
		State s = Game.getInstance().getCurrentState();
		if (Game.getInstance().getTeamA() == this){
			if (s.isReceptionStateTeamB()
					|| s.isDefenceStateTeamB()
					|| s.isSettingStateTeamB()
					|| s.isAttackingStateTeamB()){
				return true;
			}
		}
		if (Game.getInstance().getTeamB() == this){
			if (s.isReceptionStateTeamA()
					|| s.isDefenceStateTeamA()
					|| s.isSettingStateTeamA()
					|| s.isAttackingStateTeamA()){
				return true;
			}
		}
		return false;
	}

}
